#!/usr/bin/env python
# FILE: classify.py
# AUTHOR: Duong Vu
# CREATE DATE: 07 June 2021
#from sklearn.model_selection import StratifiedKFold
#from sklearn.preprocessing import StandardScaler
import sys
if sys.version_info[0] >= 3:
	unicode = str
import os, argparse
#from sklearn.metrics import precision_recall_fscore_support
#from sklearn.metrics import cohen_kappa_score
#from sklearn.metrics import matthews_corrcoef
#from sklearn.metrics import confusion_matrix
#from sklearn.metrics import accuracy_score
import json
from Bio import SeqIO
import random
import multiprocessing
parser=argparse.ArgumentParser(prog='classify.py',  
							   usage="%(prog)s [options] -i bestmatch/classified file -f the fasta file -r referencefastafilename -c classificationfile -mp minproba -mc mincoverage -cutoffs cutoffsfile -o output",
							   description='''Script that assigns the classified sequences of the prediction file to their BLAST best match based on the given cutoffs.''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--input', required=True, help='the classified file')
parser.add_argument('-fmt','--inputformat', default="tab delimited", help='the format of the classified file. The inputfmt can have two values "tab delimited" and "blast". The value "tab delimited" is given as default, and the "blast" fmt is the format of the BLAST output with outfmt=6.')
parser.add_argument('-f','--fasta', default="", help='the fasta file')
parser.add_argument('-r','--reference', default="", help='the reference fasta file')
parser.add_argument('-o','--out', default="dnabarcoder", help='The output folder.')
parser.add_argument('-c','--classification', default="", help='the classification file in tab. format.')
parser.add_argument('-mp','--minproba', type=float, default=0, help='The minimum probability for verifying the classification results.')
parser.add_argument('-ml','--minalignmentlength', type=int, default=400, help='Minimum sequence alignment length required for BLAST. For short barcode sequences like ITS2 (ITS1) sequences, minalignmentlength should probably be set to smaller, 50 for instance.')
parser.add_argument('-m','--maxseqno', type=int, default=0, help='Maximum number of the sequences of the predicted taxon name from the classification file will be selected for the comparison to find the best match. If it is not given, all the sequences will be selected.')
parser.add_argument('-cutoff','--cutoff', type=float, default=0,help='The cutoff to assign the sequences to predicted taxa. If the cutoffs file is not given, this value will be taken for sequence assignment.')
parser.add_argument('-confidence','--confidence', type=float,default=0,help='The confidence of the cutoff to assign the sequences to predicted taxa')
parser.add_argument('-rank','--classificationrank', default="", help='the classification rank')
parser.add_argument('-prefix','--prefix', help='the prefix of output filenames')
parser.add_argument('-cutoffs','--cutoffs', help='The json file containing the cutoffs to assign the sequences to the predicted taxa.')
parser.add_argument('-minseqno','--minseqno', type=int, default=0, help='the minimum number of sequences for using the predicted cut-offs to assign sequences. Only needed when the cutoffs file is given.')
parser.add_argument('-mingroupno','--mingroupno', type=int, default=0, help='the minimum number of groups for using the predicted cut-offs to assign sequences. Only needed when the cutoffs file is given.')


args=parser.parse_args()
predictionfilename=args.input
fastafilename= args.fasta
referencefastafilename= args.reference
minprobaforBlast=args.minproba
mincoverage = args.minalignmentlength
cutoff=args.cutoff
classificationfilename=args.classification
classificationrank=args.classificationrank
maxseqno=args.maxseqno
prefix=args.prefix
outputpath=args.out
cutoffsfilename=args.cutoffs
globalconfidence=args.confidence

if not os.path.exists(outputpath):
	os.system("mkdir " + outputpath)



nproc=multiprocessing.cpu_count()

def GetBase(filename):
	return filename[:-(len(filename)-filename.rindex("."))]
	
def GetWorkingBase(filename):
	basename=os.path.basename(filename)
	if "." in basename:
		basename=basename[:-(len(basename)-basename.rindex("."))] 
	path=outputpath + "/" + filename
	return path

def GetRankClassification(level,classification):
	if classification=="" or level ==-1:
		return "k__unidentified;p__unidentified;c__unidentified;o__unidentified;f__unidentified;g__unidentified;s__unidentified"	
	species=classification.split(";")[6].replace("s__","")
	genus=classification.split(";")[5].replace("g__","")
	family=classification.split(";")[4].replace("f__","")
	order=classification.split(";")[3].replace("o__","")
	bioclass=classification.split(";")[2].replace("c__","")
	phylum=classification.split(";")[1].replace("p__","")
	kingdom=classification.split(";")[0].replace("k__","")
	newclassification=""
	if level >=0 and kingdom!="unidentified":
		newclassification="k__" + kingdom +";p__unidentified;c__unidentified;o__unidentified;f__unidentified;g__unidentified;s__unidentified"	
	if level >=1 and phylum!="unidentified":
		newclassification="k__" + kingdom +";p__"+phylum +";c__unidentified;o__unidentified;f__unidentified;g__unidentified;s__unidentified"
	if level >=2 and bioclass!="unidentified":
		newclassification="k__" + kingdom +";p__"+phylum +";c__"+bioclass+";o__unidentified;f__unidentified;g__unidentified;s__unidentified"
	if level >=3 and order!="unidentified":
		newclassification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order + ";f__unidentified;g__unidentified;s__unidentified"
	if level >=4 and family!="unidentified":
		newclassification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order+";f__"+family +";g__unidentified;s__unidentified"
	if level >=5 and genus!="unidentified":
		newclassification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order+";f__"+family + ";g__"+ genus + ";s__unidentified"
	if level >=6 and species!="unidentified":
		newclassification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order+";f__"+family + ";g__"+ genus+";s__"+species 	
	return newclassification

def GetTaxonomicClassification(level,header,texts):
	classification=""
	p_s=len(texts)
	p_g=len(texts)
	p_f=len(texts)
	p_o=len(texts)
	p_c=len(texts)
	p_p=len(texts)
	p_k=len(texts)
	i=0
	for text in header.split("\t"):
		text=text.rstrip()
		if text.lower()=="species":
			p_s=i
		elif text.lower()=="genus":
			p_g=i	
		elif text.lower()=="family":
			p_f=i	
		elif text.lower()=="order":
			p_o=i	
		elif text.lower()=="class":
			p_c=i	
		elif text.lower()=="phylum":
			p_p=i	
		elif text.lower()=="kingdom":
			p_k=i	
		i=i+1 
	species="unidentified"
	genus="unidentified"
	family="unidentified"
	order="unidentified" 
	bioclass="unidentified"
	phylum="unidentified"
	kingdom="unidentified"
	if p_s< len(texts):
		species=texts[p_s].rstrip()
		if species=="" or ("unculture" in species) or ("_sp_" in species) or ("unidentified" in species) or ("environmental" in species.lower()) or ("(" in species):
			species="unidentified"
		elif not (" " in species or "_" in species):
			species="unidentified"
		species=species.replace(" ","_")
	if p_g< len(texts):
		genus=texts[p_g].rstrip()
		if genus=="" or ("unculture" in genus) or ("unidentified" in genus):
			genus="unidentified"
	if p_f< len(texts):
		family=texts[p_f].rstrip()
		if family=="" or ("unculture" in family) or ("unidentified" in family):
			family="unidentified"
	if p_o< len(texts):
		order=texts[p_o].rstrip()
		if order=="" or ("unculture" in order) or ("unidentified" in order):
			order="unidentified"
	if p_c< len(texts):
		bioclass=texts[p_c].rstrip()
		if bioclass=="" or ("unculture" in bioclass) or ("unidentified" in bioclass):
			bioclass="unidentified"
	if p_p< len(texts):
		phylum=texts[p_p].rstrip()
		if phylum=="" or ("unculture" in phylum) or ("unidentified" in phylum):
			phylum="unidentified"
	if p_k< len(texts):
		kingdom=texts[p_k].rstrip()
		if kingdom=="" or ("unculture" in kingdom) or ("unidentified" in kingdom):
			kingdom="unidentified"
	taxonname=""
	rank=""
	if level <7 and kingdom!="unidentified":
		taxonname=kingdom
		rank="kingdom"
		classification="k__" + kingdom +";p__unidentified;c__unidentified;o__unidentified;f__unidentified;g__unidentified;s__unidentified"	
	if level <6 and phylum!="unidentified":
		taxonname=phylum
		rank="phylum"
		classification="k__" + kingdom +";p__"+phylum +";c__unidentified;o__unidentified;f__unidentified;g__unidentified;s__unidentified"
	if level <5 and bioclass!="unidentified":
		taxonname=bioclass
		rank="class"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass+";o__unidentified;f__unidentified;g__unidentified;s__unidentified"
	if level <4 and order!="unidentified":
		taxonname=order
		rank="order"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order + ";f__unidentified;g__unidentified;s__unidentified"
	if level <3 and family!="unidentified":
		taxonname=family
		rank="family"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order+";f__"+family +";g__unidentified;s__unidentified"
	if level <2 and genus!="unidentified":
		taxonname=genus
		rank="genus"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order+";f__"+family + ";g__"+ genus + ";s__unidentified"
	if level <1 and species!="unidentified":
		taxonname=species
		rank="species"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order+";f__"+family + ";g__"+ genus+";s__"+species 	
	return classification,taxonname,rank

def LoadClassification(seqrecords,classificationfilename):
	classificationdict={}
	classes={}
	if classificationfilename == "":
		return {},{}
	classificationfile= open(classificationfilename)
	header=next(classificationfile)
	for line in classificationfile:
		texts=line.split("\t")
		seqid = texts[0].replace(">","").rstrip()
		classification,taxonname,rank=GetTaxonomicClassification(0,header,texts)
		if classification!="":
			classificationdict.setdefault(seqid,{})
			classificationdict[seqid]["classification"]=classification
			classificationdict[seqid]["taxonname"]=taxonname
			classificationdict[seqid]["rank"]=rank	
		if seqid in seqrecords.keys():
			taxonnames=classification.split(";")
			for taxonname in taxonnames:
				if taxonname=="":
					continue
				taxonname=taxonname.split("__")[1].replace("_"," ")
				if taxonname=="unidentified" or taxonname=="":
					continue
				if taxonname in classes.keys():
					classes[taxonname].append(seqrecords[seqid])
				else:
					classes.setdefault(taxonname,[seqrecords[seqid]])		
	classificationfile.close()	
	return classificationdict,classes

#def LoadTaxa(classificationfilename):
#	taxa=[]
#	classificationfile=open(classificationfilename)
#	for line in classificationfile:
#		line=line.rstrip()
#		texts=line.split("\t")
#		for text in texts:
#			if text!="":
#				taxa.append(text)
#	classificationfile.close()			
#	return taxa

def CreateFastaFile(taxonname,classeswithsequences,maxseqno):
	if sys.version_info[0] < 3:
		taxonname=unicode(taxonname,errors='ignore')
	newfastafilename=""
	sequences=[]
	if taxonname in classeswithsequences.keys():
		sequences=classeswithsequences[taxonname]
		if len(sequences) >0:
			newfastafilename=taxonname.replace(" ","_").replace(".","_") + ".fasta"
			fastafile=open(newfastafilename,"w")
			if (maxseqno >0) and (len(sequences) > maxseqno):
				#select randomly 100 sequences to compare
				selectedlist=random.sample(range(0, len(sequences)), k=maxseqno)
				for i in selectedlist:
					sequence=sequences[i]
					seqid=sequence.id
					seq=str(sequence.seq)
					fastafile.write(">" + seqid + "\n")
					fastafile.write(seq + "\n")
			else:	
				for sequence in sequences:
					seqid=sequence.id
					seq=str(sequence.seq)
					fastafile.write(">" + seqid + "\n")
					fastafile.write(seq + "\n")
			fastafile.close()
	return newfastafilename,len(sequences)

def ComputeBestLocalBLASTScore(testrecord,reffilename,mincoverage):
	#Create fasta file of the test record 
	queryname=testrecord.id + "." + "fasta"
	if "|" in testrecord.id:
		queryname=testrecord.id.split("|")[0] + "." + "fasta"
	SeqIO.write(testrecord, queryname , "fasta")
	#Create fasta file of predictedname
	if reffilename=="":
		return "",0,0,0
	#Blast the test record to fastafilename:
	#makedbcommand = "makeblastdb -in " + refname + " -dbtype \'nucl\' " +  " -out db"
	#os.system(makedbcommand)
	blastoutput = testrecord.id 
	if "|" in testrecord.id:
		blastoutput=blastoutput.split("|")[0]
	#blast
	makedbcommand = "makeblastdb -in " + reffilename + " -dbtype \'nucl\' " +  " -out db"
	os.system(makedbcommand)
	
	blastoutput= blastoutput + ".blast.out"
	#blastcommand = "blastn -query " + queryname + " -db  db -outfmt 6 -out " + blastoutput
	blastcommand = "blastn -query " + queryname + " -db  db -task blastn-short -outfmt 6 -out " + blastoutput + " -num_threads " + str(nproc)
	#for long read
	if mincoverage >=300:
		blastcommand = "blastn -query " + queryname + " -db  db -outfmt 6 -out " + blastoutput + " -num_threads " + str(nproc)
	os.system(blastcommand)
	#read output of Blast
	blastoutputfile = open(blastoutput)
	bestlocalscore=0
	bestlocalsim=0
	bestlocalcoverage=0
	bestrefid=""
	for line in blastoutputfile:
		words = line.split("\t")
		pos1 = int(words[6])
		pos2 = int(words[7])
		iden = float(words[2]) 
		sim=float(iden)/100
		coverage=abs(pos2-pos1)
		refid=words[1]
		score=sim
		if coverage < mincoverage:
				score=float(score * coverage)/mincoverage
		if score > bestlocalscore:
			bestrefid=refid
			bestlocalscore=score
			bestlocalsim=sim
			bestlocalcoverage=coverage
	os.system("rm " + blastoutput)
	os.system("rm " + queryname)
	return bestrefid,bestlocalscore,bestlocalsim,bestlocalcoverage

def GetLevel(rank):
	level=-1
	if rank=="species":	
		level=6
	elif rank=="genus":	
		level=5
	elif rank=="family":	
		level=4
	elif rank=="order":	
		level=3
	elif rank=="class":	
		level=2
	elif rank=="phylum":
		level=1
	elif rank=="kingdom":
		level=0	
	return level


def GetHigherTaxa(rank,classification):
	highertaxa=[]
	taxa=classification.split(";")
	species=""
	genus=""
	family=""
	order=""
	bioclass=""
	phylum=""
	kingdom=""
	for taxon in taxa:
		if taxon.startswith("k__"):
			kingdom=taxon.replace("k__","")
		if taxon.startswith("p__"):
			phylum=taxon.replace("p__","")	
		if taxon.startswith("c__"):
			bioclass=taxon.replace("c__","")	
		if taxon.startswith("o__"):
			order=taxon.replace("o__","")	
		if taxon.startswith("f__"):
			family=taxon.replace("f__","")	
		if taxon.startswith("g__"):
			genus=taxon.replace("g__","")	
		if taxon.startswith("s__"):
			species=taxon.replace("s__","")	
	level=GetLevel(rank)	
	if level >=6 and species!="" and species!="unidentified":
		highertaxa.append(species)
	if level >=5 and genus!="" and genus!="unidentified":
		highertaxa.append(genus)
	if level >=4 and family!="" and family!="unidentified":
		highertaxa.append(family)	
	if level >=3 and order!="" and order!="unidentified":
		highertaxa.append(order)	
	if level >=2 and bioclass!="" and bioclass!="unidentified":
		highertaxa.append(bioclass)	
	if level >=1 and phylum!="" and phylum!="unidentified":
		highertaxa.append(phylum)
	if level >=0 and kingdom!="" and kingdom!="unidentified":
		highertaxa.append(kingdom)
	return highertaxa

def GetCutoffAndConfidence(rank,classification,cutoffs):
	if not rank in cutoffs.keys():
		return [0,0]
	#cleanclassification=classification.replace("k__","").replace("p__","").replace("c__","")	.replace("o__","").replace("f__","").replace("g__","").replace("s__","")
	#taxa=cleanclassification.split(";")
	#taxa.append("All") 
	highertaxa=GetHigherTaxa(rank,classification)
	highertaxa.append("All") 
	localcutoff=0
	seqno=0
	groupno=0
	datasets=cutoffs[rank]
	maxconfidence=0
	bestcutoff=0
	for highertaxonname in highertaxa:
		if not highertaxonname in datasets.keys():
			continue
		if "cut-off" in datasets[highertaxonname].keys():
			localcutoff=datasets[highertaxonname]["cut-off"]
		else:
			continue
		confidence=1	
		if "confidence" in datasets[highertaxonname].keys():
			confidence=datasets[highertaxonname]["confidence"]
		if "sequence number" in datasets[highertaxonname].keys():
			seqno=datasets[highertaxonname]["sequence number"]	
		if "group number" in datasets[highertaxonname].keys():
			groupno=datasets[highertaxonname]["group number"]	
		if not ((seqno >0 and seqno < args.minseqno) or (groupno >0 and groupno < args.mingroupno)):	
			if maxconfidence < confidence:
				maxconfidence =confidence
				bestcutoff=localcutoff
				break
	return [bestcutoff,maxconfidence]

def GetCutoffs(classification,cutoffs):
	species="unindentified"
	genus="unindentified"
	family="unindentified"
	order="unindentified"
	bioclass="unindentified"
	phylum="unindentified"
	kingdom="unindentified"
	texts=classification.split(";")
	for text in texts:
		if text.startswith("s__"):
			species=text.replace("s__","")
			species=species.replace("_"," ")
		if text.startswith("g__"):
			genus=text.replace("g__","")	
		if text.startswith("f__"):
			family=text.replace("f__","")	
		if text.startswith("o__"):
			order=text.replace("o__","")	
		if text.startswith("c__"):
			bioclass=text.replace("c__","")	
		if text.startswith("p__"):
			phylum=text.replace("p__","")	
		if text.startswith("k__"):
			kingdom=text.replace("k__","")	
	taxacutoffs={}		
	taxacutoffs["species"]=GetCutoffAndConfidence("species",classification,cutoffs)
	taxacutoffs["genus"]=GetCutoffAndConfidence("genus",classification,cutoffs)
	taxacutoffs["family"]=GetCutoffAndConfidence("family",classification,cutoffs)
	taxacutoffs["order"]=GetCutoffAndConfidence("order",classification,cutoffs)
	taxacutoffs["class"]=GetCutoffAndConfidence("class",classification,cutoffs)
	taxacutoffs["phylum"]=GetCutoffAndConfidence("phylum",classification,cutoffs)
	taxacutoffs["kingdom"]=GetCutoffAndConfidence("kingdom",classification,cutoffs)
	return taxacutoffs,kingdom,phylum,bioclass,order,family,genus,species

def GetAssignment(refid,classificationdict,bestscore,cutoffs,cutoff,globalconfidence,classificationrank):
	confidence=globalconfidence
	rank=classificationrank
	localcutoff=cutoff
	taxonname=""
	level=-1
	classification=""
	if cutoffs!={} and refid!="" and (refid in classificationdict.keys()):
		refclassification=classificationdict[refid]['classification']
		taxacutoffs,kingdom,phylum,bioclass,order,family,genus,species=GetCutoffs(refclassification,cutoffs)
		if bestscore >=taxacutoffs["species"][0] and taxacutoffs["species"][0] >0 and species!="unidentified" and (classificationrank=="species" or classificationrank== ""):
			rank="species"
			localcutoff=taxacutoffs["species"][0]
			confidence=taxacutoffs["species"][1]
			taxonname=species		
			level=6
		elif bestscore >=taxacutoffs["genus"][0] and taxacutoffs["genus"][0]>0 and genus!="unidentified" and (classificationrank=="genus" or classificationrank== ""):
			rank="genus"
			localcutoff=taxacutoffs["genus"][0]
			confidence=taxacutoffs["genus"][1]
			taxonname=genus
			level=5
		elif bestscore >=taxacutoffs["family"][0] and taxacutoffs["family"][0] >0 and family!="unidentified" and (classificationrank=="family" or classificationrank== ""):
			rank="family"
			localcutoff=taxacutoffs["family"][0]
			confidence=taxacutoffs["family"][1]
			taxonname=family	
			level=4
		elif bestscore >=taxacutoffs["order"][0] and taxacutoffs["order"][0] >0 and order!="unidentified" and (classificationrank=="order" or classificationrank== ""):
			rank="order"
			localcutoff=taxacutoffs["order"][0]
			confidence=taxacutoffs["order"][1]
			taxonname=order
			level=3
		elif bestscore >=taxacutoffs["class"][0] and taxacutoffs["class"][0] >0 and bioclass!="unidentified" and (classificationrank=="class" or classificationrank== ""):
			rank="class"
			localcutoff=taxacutoffs["class"][0]
			confidence=taxacutoffs["class"][1]
			taxonname=bioclass
			level=2
		elif bestscore >=taxacutoffs["phylum"][0] and taxacutoffs["phylum"][0]>0 and phylum!="unidentified" and (classificationrank=="phylum" or classificationrank== ""):
			rank="phylum"
			localcutoff=taxacutoffs["phylum"][0]
			confidence=taxacutoffs["phylum"][1]
			taxonname=phylum
			level=1
		elif bestscore >=taxacutoffs["kingdom"][0] and taxacutoffs["kingdom"][0]>0 and kingdom!="unidentified" and (classificationrank=="kingdom" or classificationrank== ""):
			rank="kingdom"
			localcutoff=taxacutoffs["kingdom"][0]
			confidence=taxacutoffs["kingdom"][1]
			taxonname=kingdom
			level=0	
		level=GetLevel(rank)
		classification=GetRankClassification(level,refclassification)
	elif refid!="" and (refid in classificationdict.keys()):
		refclassification=classificationdict[refid]['classification']
		if bestscore >=cutoff:
			if classificationrank=="":
				taxonname=classificationdict[refid]['taxonname']
				rank=classificationdict[refid]['rank']
				level=GetLevel(rank)	
			else:	
				level=GetLevel(classificationrank)	
				rank=classificationrank
				taxonname=(refclassification.replace("k__","").replace("p__","").replace("c__","")	.replace("o__","").replace("f__","").replace("g__","").replace("s__","")).split(";")[level]
		classification=GetRankClassification(level,refclassification)
	else:
		classification=GetRankClassification(-1,classification)
	return classification,taxonname,rank,level,localcutoff,confidence

def Assign(classeswithsequences,refclassificationdict,queryclassificationdict,predictedclassificationdict,minprobaforBlast,mincoverage,cutoffs,cutoff,globalconfidence,seqdict,bestmatchdict,maxseqno,outputname,classificationreportfilename):
	#classificationlevel=GetLevel(classificationrank)
	output=open(outputname,"w")
	classificationreportfile=open(classificationreportfilename,"w")
	output.write("#SequenceID\tGiven label\tPrediction\tFull classification\tProbability\tRank\tCut-off\tConfidence\tReferenceID\tBLAST score\tBLAST sim\tBLAST coverage\n")
	classificationreportfile.write("SequenceID\tReferenceID\tkingdom\tphylum\tclass\torder\tfamily\tgenus\tspecies\trank\tscore\tcutoff\tconfidence\n")
	i=0
	given_labels=[]
	assigned_labels=[]
	unclassifiedseqids=[]
	count=0
	for seqid in bestmatchdict.keys():
		proba =1
		rank=""
		if "proba" in bestmatchdict[seqid].keys():
			proba=bestmatchdict[seqid]["proba"]
		predictedname= ""
		if "predlabel" in bestmatchdict[seqid].keys():
			predictedname=bestmatchdict[seqid]["predlabel"]
		giventaxonname=""
		if "label" in bestmatchdict[seqid].keys():
			giventaxonname=bestmatchdict[seqid]["label"]
		classification=""
		if "predclassification" in bestmatchdict[seqid].keys():
			classification=bestmatchdict[seqid]["predclassification"]
		refid=bestmatchdict[seqid]["refid"]
		bestscore=bestmatchdict[seqid]["score"]
		sim=bestmatchdict[seqid]["sim"]
		coverage=bestmatchdict[seqid]["alignmentlength"]
		if  proba >= minprobaforBlast:
			if refid=="":
				if seqid in seqdict.keys():
					reffilename,numberofsequences=CreateFastaFile(predictedname,classeswithsequences,maxseqno)
					if reffilename!="":
						newrefid,newbestscore,newsim,newcoverage=ComputeBestLocalBLASTScore(seqdict[seqid],reffilename,mincoverage)
						os.system("rm " + reffilename)	
						if newbestscore > bestscore:
							refid=newrefid
							bestscore=newbestscore
							sim=newsim
							coverage=newcoverage
			if refid!="":
				classification,predictedname,rank,level,cutoff,confidence=GetAssignment(refid,refclassificationdict,bestscore,cutoffs,cutoff,globalconfidence,classificationrank)			
			else:
				classification,predictedname,rank,level,cutoff,confidence=GetAssignment(seqid,predictedclassificationdict,bestscore,cutoffs,cutoff,globalconfidence,classificationrank)			
			if sys.version_info[0] < 3:
				predictedname=unicode(predictedname,'latin1')
		if seqid in refclassificationdict.keys():
			giventaxa=refclassificationdict[seqid]['classification']
			giventaxa=giventaxa.replace("k__","").replace("p__","").replace("c__","")	.replace("o__","").replace("f__","").replace("g__","").replace("s__","")
			giventaxonname=giventaxa.split(";")[level]		
		if giventaxonname!="" and giventaxonname!="unidentified":	
			giventaxonname=giventaxonname.replace("_"," ")
			predictedname=predictedname.replace("_"," ")
			given_labels.append(giventaxonname)
			assigned_labels.append(predictedname)	
#		if predictedname!="" and predictedname!="unidentified":			
#			output.write(seqid + "\t" + giventaxonname + "\t"  + predictedname + "\t"+ classification + "\t" + str(proba) + "\t" + rank + "\t" + str(cutoff) + "\t" + str(confidence) + "\t" + refid + "\t" + str(bestscore) + "\t" + str(sim) + "\t" + str(coverage) + "\n")			
#			cleanclassification=classification.replace("k__","").replace("p__","").replace("c__","")	.replace("o__","").replace("f__","").replace("g__","").replace("s__","").replace("_"," ")
#			if refid==seqid:
#				refid=""
#			classificationreportfile.write(seqid + "\t" + refid + "\t" + cleanclassification.replace(";","\t") + "\t" + rank + "\t" + str(bestscore) + "\t" + str(cutoff) + "\t" + str(confidence) + "\n")
#			count=count+1
#		else:
#			unclassifiedseqids.append(seqid)	
		#save all classification
		if predictedname!="" and predictedname!="unidentified":
			count=count+1
		else:
			unclassifiedseqids.append(seqid)	
			rank=""
		output.write(seqid + "\t" + giventaxonname + "\t"  + predictedname + "\t"+ classification + "\t" + str(proba) + "\t" + rank + "\t" + str(cutoff) + "\t" + str(confidence) + "\t" + refid + "\t" + str(bestscore) + "\t" + str(sim) + "\t" + str(coverage) + "\n")			
		cleanclassification=classification.replace("k__","").replace("p__","").replace("c__","")	.replace("o__","").replace("f__","").replace("g__","").replace("s__","").replace("_"," ")
#		if refid==seqid:
#			refid=""
		classificationreportfile.write(seqid + "\t" + refid + "\t" + cleanclassification.replace(";","\t") + "\t" + rank + "\t" + str(bestscore) + "\t" + str(cutoff) + "\t" + str(confidence) + "\n")
		
	output.close()
	classificationreportfile.close()
	return count,given_labels,assigned_labels,unclassifiedseqids
	
def LoadPrediction(predictionfilename,mincoverage):
	bestmatchdict={}
	p_id=-1
	p_l=-1
	p_pl=-1
	p_c=-1
	p_p=-1
	p_r=-1
	p_bs=-1
	p_s=-1
	p_co=-1
	predictionfile=open(predictionfilename)
	headers=next(predictionfile)
	i=0
	for header in headers.rstrip().split("\t"):
		if ("sequenceid" in header.lower()) or ("sequence_id" in header.lower()) or ("sequence id" in header.lower()) or ("seqid" in header.lower()) :
			p_id=i
		if "given label" in header.lower():
			p_l=i
		if "prediction" in header.lower():
			p_pl=i
		if "full classification" in header.lower():
			p_c=i
		if "probability" in header.lower():
			p_p=i
		if ("referenceid" in header.lower()) or ("reference_id" in header.lower()) or ("reference id" in header.lower()) or ("refid" in header.lower()) :
			p_r=i
		if "score" in header.lower():
			p_bs=i	
		if "sim" in header.lower():
			p_s=i
		if "coverage" in header.lower():
			p_co=i	
		i=i+1	
	for line in predictionfile:
		texts=line.rstrip().split("\t")
		seqid=""
		if p_id >=0 and p_id < len(texts):
			seqid=texts[p_id]
		if seqid=="":
			continue
		label=""
		if p_l >=0 and p_l < len(texts):
			label=texts[p_l]
		predlabel=""
		if p_pl >=0 and p_pl < len(texts):
			predlabel=texts[p_l]
		predclassification=""
		if p_c >=0 and p_c < len(texts):
			predclassification=texts[p_c]
		proba=1
		if p_p >=0 and p_p <len(texts):
			proba=float(texts[p_p])
		refid=""
		if p_r>0 and p_r<len(texts):
			refid=texts[p_r]
		sim=0
		if p_s>0 and p_s<len(texts):
			sim=float(texts[p_s])
		alignmentlength=""
		if p_co>0 and p_co<len(texts):
			alignmentlength=float(texts[p_co])
		score=0
		if p_bs>0 and p_bs<len(texts):
			score=float(texts[p_bs])
		else:
			score=sim
			if alignmentlength < mincoverage:
				score=float(score * alignmentlength)/mincoverage
		bestmatchdict.setdefault(seqid,{})
		bestmatchdict[seqid]["refid"]=refid
		bestmatchdict[seqid]["label"]=label
		bestmatchdict[seqid]["predlabel"]=predlabel
		bestmatchdict[seqid]["predclassification"]=predclassification
		bestmatchdict[seqid]["proba"]=proba
		bestmatchdict[seqid]["score"]=score
		bestmatchdict[seqid]["sim"]=sim
		bestmatchdict[seqid]["alignmentlength"]=alignmentlength
	return bestmatchdict

def LoadBlastOutput(blastoutput,mincoverage):
	bestmatchdict={}
	#read blast output
	blastoutputfile = open(blastoutput)
	refid = ""
	score=0
	seqid=""
	for line in blastoutputfile:
		words = line.split("\t")
		seqid=words[0]
		pos1 = int(words[6])
		pos2 = int(words[7])
		iden = float(words[2]) 
		sim=float(iden)/100
		coverage=abs(pos2-pos1)
		refid=words[1]
		score=sim
		if coverage < mincoverage:
			score=float(score * coverage)/mincoverage
		if not (seqid in bestmatchdict.keys()):
			bestmatchdict.setdefault(seqid,{})
			bestmatchdict[seqid]["refid"]=""
			bestmatchdict[seqid]["score"]=0
			bestmatchdict[seqid]["sim"]=0
			bestmatchdict[seqid]["alignmentlength"]=0
		#if (score > bestmatchdict[seqid]["score"]):
		if (score > bestmatchdict[seqid]["score"]) or (score == bestmatchdict[seqid]["score"] and coverage > bestmatchdict[seqid]["alignmentlength"]):
			bestmatchdict[seqid]["refid"]=refid
			bestmatchdict[seqid]["score"]=score
			bestmatchdict[seqid]["sim"]=sim
			bestmatchdict[seqid]["alignmentlength"]=coverage
	return bestmatchdict
	
def GetClassificationpos(pred_labels,classificationfilename):
	classificationpos=0
	rank=""
	classificationfile=open(classificationfilename)
	for line in classificationfile:
		if line.startswith("#"):
			continue
		texts=line.rstrip().split("\t")	 
		i=0
		for text in texts:
			if text!="":
				if text in pred_labels:
					classificationpos=i
					break
			i=i+1	
		if classificationpos >0:
			break
	classificationfile.close()	
	return classificationpos,rank

def LoadClassificationForKronaReport(classificationfilename):
	classificationdict={}
	classificationfile= open(classificationfilename)
	next(classificationfile)
	for line in classificationfile:
		if line.startswith("#"):
			continue 
		texts=line.split("\t")
		classification=texts[3]
		classification=classification.replace("k__","").replace("p__","").replace("c__","").replace("o__","").replace("f__","").replace("g__","").replace("s__","")
		if ";" in classification:
			texts=classification.split(";")
			classification=texts[0]
			if classification=="":
				classification="unidentified"
			for i in range(1,len(texts)):
				text=texts[i]
				if text=="":
					text="unidentified"
				classification=classification + "\t" + text	
		if classification in classificationdict.keys():
			classificationdict[classification]=classificationdict[classification] + 1
		else:
			classificationdict.setdefault(classification, 1)
	classificationfile.close()	
	return classificationdict

def mergeDict(dict1,dict2):
	dict1.update(dict2)
	for key, value in dict1.items():
		if key in dict1 and key in dict2:
			dict1[key] = value + dict2[key]
	return dict1

def KronaPieCharts(classification,kronareport,kronahtml):
	kronareportfile=open(kronareport,"w")
	for classname in classification.keys():
		kronareportfile.write(str(classification[classname]) + "\t" + classname + "\n")
	kronareportfile.close()	
	#create kronahtml
	command="ImportText.pl " + kronareport + " -o " + kronahtml
	#print(command)
	os.system(command)
	os.system("firefox " + kronahtml) 
	
if __name__ == "__main__":
	if prefix=="" or prefix==None:
		prefix=GetBase(predictionfilename)
		if "/" in prefix:
			prefix=prefix[prefix.rindex("/")+1:]	
		if cutoff >0:
			prefix =prefix + "." + str(cutoff)
	outputname=GetWorkingBase(prefix) + ".classified"
	if classificationrank!="":
		outputname=GetWorkingBase(prefix) + "." + classificationrank + ".classified"
	if outputname==predictionfilename:
		outputname=outputname+".classified"
	classificationreportfilename=GetBase(outputname) + ".classification"
	unclassifiedfastafilename=GetBase(outputname)  + ".unclassified.fasta"	
	#load prediction
	bestmatchdict={}
	if args.inputformat=="blast":
		bestmatchdict=LoadBlastOutput(predictionfilename,mincoverage)	
	else:
		bestmatchdict=LoadPrediction(predictionfilename,mincoverage)	
	#load sequences
	seqrecords={}
	if os.path.exists(fastafilename):
		seqrecords=SeqIO.to_dict(SeqIO.parse(fastafilename, "fasta"))
	#load reference sequences:	
	refseqrecords={}
	if os.path.exists(referencefastafilename):
		refseqrecords=SeqIO.to_dict(SeqIO.parse(referencefastafilename, "fasta"))
	refclassificationdict,refclasses= LoadClassification(refseqrecords,classificationfilename)
	queryclassificationdict,queryclasses= LoadClassification(seqrecords,classificationfilename)	
	predictedclassificationdict,predictedclasses= LoadClassification(seqrecords,predictionfilename)	
	cutoffs={}
	if cutoffsfilename!="" and cutoffsfilename!=None:
		with open(cutoffsfilename) as cutoffsfile:
			cutoffs = json.load(cutoffsfile)		
	count,given_labels,assigned_labels,unclassifiedseqids=Assign(refclasses,refclassificationdict,queryclassificationdict,predictedclassificationdict,minprobaforBlast,mincoverage,cutoffs,cutoff,globalconfidence,seqrecords,bestmatchdict,maxseqno,outputname,classificationreportfilename)
	print("Number of classified sequences: " + str(count))
	#print("The results are saved in file  " + outputname)
	print("The results are saved in file  " + outputname + " and " + classificationreportfilename + ".")
	unclassifiedseqrecords=[]
	if len(unclassifiedseqids) > 0:
		for seqid in unclassifiedseqids:
			if seqid in seqrecords.keys():
				unclassifiedseqrecords.append(seqrecords[seqid])
		#write to fasta file	
		if len(unclassifiedseqrecords)>0:
			SeqIO.write(unclassifiedseqrecords, unclassifiedfastafilename, "fasta")	
			print("The unclassified sequences are saved in the file " +   unclassifiedfastafilename + ".")
	
#	if len(given_labels) >0:
#		reportname=GetBase(outputname) + ".report"
#		reftaxa=LoadTaxa(classificationfilename)
#		CalculateClassificationMetrics(given_labels,pred_labels,reftaxa,reportname)
	#Compute classification metrices
	#making krona report
	if count > 0:
		kronareport = GetBase(outputname) + ".krona.report"
		kronahtml=GetBase(kronareport) + ".html"
		classificationdict= LoadClassificationForKronaReport(outputname)
		KronaPieCharts(classificationdict,kronareport,kronahtml)
		print("The krona report and html are saved in files " + kronareport + " and " + kronahtml + ".") 



#!/usr/bin/env python
# FILE: removesequences.py
# AUTHOR: Duong Vu
# CREATE DATE: 07 June 2020
import sys
import numpy as np
import os, argparse
from Bio import SeqIO
import multiprocessing
nproc=multiprocessing.cpu_count()
#from keras.utils import np_utils

parser=argparse.ArgumentParser(prog='removesequences.py', 
							   usage="%(prog)s [options] -i fastafile -c classificationfile -t taxa -o output -rank species",
							   description='''Script that selects the sequences for of the given taxa. The taxon names in the taxa are separated by ","''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--input', required=True, help='the fasta file')
parser.add_argument('-o','--out', help='The fasta output file containing the sequences of the given taxa.') #optional
parser.add_argument('-c','--classification', required=True, help='the classification file in tab. format.')
parser.add_argument('-t','--taxa', default="", help='the taxa for the selection, separated by ",". If taxa="", all the sequences with the ids given in the classification file will be removed from the input file.')
parser.add_argument('-rank','--classificationrank', default="", help='If the value of the classification rank is not empty, the sequences will be removed.')
parser.add_argument('-seqidpos','--sequenceidposition', type=int,default=0, help='the position of sequence id in the classification file.')

args=parser.parse_args()
fastafilename= args.input
classificationfilename=args.classification
taxa=args.taxa
output=args.out
classificationrank=args.classificationrank
seqidpos=args.sequenceidposition

#fastafilename=sys.argv[1]
#taxa=sys.argv[2]
#classificationfilename = sys.argv[3]
#output=sys.argv[4]

def GetBase(filename):
	return filename[:-(len(filename)-filename.rindex("."))]

def SelectSeqIds(classificationfilename,taxa,classificationpos,seqidpos):
	if not os.path.exists(classificationfilename):
		return [],[]
	taxalist=[]
	classnames=[]
	if "," in taxa:
		taxalist=taxa.split(",")
	elif taxa!="" and taxa!="unidentified":
		taxalist.append(taxa)
	classificationfile=open(classificationfilename)
	seqids=[]
	for line in classificationfile:
		elements=line.rstrip().split("\t")
		seqid = elements[seqidpos].replace(">","").rstrip()
		classname=""
		if classificationpos >=0 and classificationpos < len(elements):
			classname=elements[classificationpos]
		if classname=="" or classname=="unidentified":
			continue
		if taxa!="":
			for taxonname in taxalist:
				if taxonname!="" and taxonname!="unidentified":
					if taxonname in elements:
						seqids.append(seqid)				
						classnames.append(classname)
		else:
			seqids.append(seqid)
			classnames.append(classname)
	return seqids,classnames

def GetPosition(classificationfilename,rank):
	pos=-1
	classificationfile=open(classificationfilename)
	header=classificationfile.readline()
	header=header.rstrip()
	classificationfile.close()
	texts=header.split("\t")
	isError=False
	if rank in texts:
		pos=texts.index(rank)
	else:
		print("The rank " + rank + " is not given in the classification." )
		isError=True
	return pos,isError

#####main###
classificationpos=-1
if classificationfilename!="":
	classificationpos,isError=GetPosition(classificationfilename,classificationrank)	
seqids,classnames=SelectSeqIds(classificationfilename,taxa,classificationpos,seqidpos)
seqrecords=list(SeqIO.parse(fastafilename, "fasta"))
selectedrecords=[]
for seqrec in seqrecords:
	seqid=seqrec.id
	if not (seqid in seqids):
		selectedrecords.append(seqrec)
#save to file:
SeqIO.write(selectedrecords,output,"fasta")
print("The selected sequences are saved in " + output + ".")




#!/usr/bin/env python
# FILE: fasta2matrix.py
# AUTHOR: William Stafford Noble
# CREATE DATE: 27 September 2005
import sys
import math
import numpy as np
#############################################################################
def make_kmer_list (k, alphabet):

    # Base case.
    if (k == 1):
        return(alphabet)

    # Handle k=0 from user.
    if (k == 0):
        return([])

    # Error case.
    if (k < 1):
        sys.stderr.write("Invalid k=%d" % k)
        sys.exit(1)

    # Precompute alphabet length for speed.
    alphabet_length = len(alphabet)

    # Recursive call.
    return_value = []
    for kmer in make_kmer_list(k-1, alphabet):
        for i_letter in range(0, alphabet_length):
            return_value.append(kmer + alphabet[i_letter])
              
    return(return_value)

##############################################################################
def make_upto_kmer_list (k_values,
                         alphabet):

    # Compute the k-mer for each value of k.
    return_value = []
    for k in k_values:
        return_value.extend(make_kmer_list(k, alphabet))

    return(return_value)
                 
##############################################################################
def normalize_vector (normalize_method,
                      k_values,
                      vector,
                      kmer_list):

    # Do nothing if there's no normalization.
    if (normalize_method == "none"):
        return(vector)

    # Initialize all vector lengths to zeroes.
    vector_lengths = {}
    for k in k_values:
        vector_lengths[k] = 0

    # Compute sum or sum-of-squares separately for each k.
    num_kmers = len(kmer_list)
    for i_kmer in range(0, num_kmers):
        kmer_length = len(kmer_list[i_kmer])
        count = vector[i_kmer]
        if (normalize_method == "frequency"):
            vector_lengths[kmer_length] += count
        elif (normalize_method == "unitsphere"):
            vector_lengths[kmer_length] += count * count

    # Square root each sum if we're doing 2-norm.
    if (normalize_method == "unitsphere"):
        for k in k_values:
            vector_lengths[k] = math.sqrt(vector_lengths[k])

    # Divide through by each sum.
    return_value = []
    for i_kmer in range(0, num_kmers):
        kmer_length = len(kmer_list[i_kmer])
        count = vector[i_kmer]
        vector_length = vector_lengths[kmer_length]
        if (vector_length == 0):
            return_value.append(0)
        else:
            return_value.append(float(count) / float(vector_length))

    return(return_value)

##############################################################################
# Make a copy of a given string, substituting one letter.
def substitute (position,
                letter,
                string):

    return_value = ""
    if (position > 0):
        return_value = return_value + string[0:position]
    return_value = return_value + letter
    if (position < (len(string) - 1)):
        return_value = return_value + string[position+1:]
                   
    return(return_value)


##############################################################################
def compute_bin_num (num_bins,
                     position,
                     k,
                     numbers):
  
  # If no binning, just return.
  if (num_bins == 1):
    return(0)

  # Compute the mean value for this k-mer.
  mean = 0
  for i in range(0, k):
    mean += float(numbers[position + i])
  mean /= k

  # Find what quantile it lies in.
  for i_bin in range(0, num_bins):
    if (mean <= boundaries[k][i_bin]):
      break

  # Make sure we didn't go too far.
  if (i_bin == num_bins):
      sys.stderr.write("bin=num_bins=%d\n", i_bin)
      sys.exit(1);

  return(i_bin)

##############################################################################
def make_sequence_vector (sequence,
                          numbers,
                          num_bins,
                          revcomp,
                          revcomp_dictionary,
                          normalize_method,
                          k_values,
                          mismatch,
                          alphabet,
                          kmer_list,
                          boundaries,
                          pseudocount):

    # Make an empty counts vector.
    kmer_counts = []
    for i_bin in range(0, num_bins):
        kmer_counts.append({})

    # Iterate along the sequence.
    for k in k_values:
        seq_length = len(sequence) - k + 1
        for i_seq in range(0, seq_length):

            # Compute which bin number this goes in.
            bin_num = compute_bin_num(num_bins, i_seq, k, numbers)

            # Extract this k-mer.
            kmer = sequence[i_seq : i_seq + k]

            # If we're doing reverse complement, store the count in the
            # the version that starts with A or C.
            if (revcomp == 1):
                rev_kmer = find_revcomp(kmer, revcomp_dictionary)
                if (cmp(kmer, rev_kmer) > 0):
                    kmer = rev_kmer

            # Increment the count.
            #if (kmer_counts[bin_num].has_key(kmer)):
            if (kmer in kmer_counts[bin_num].keys()):				
                kmer_counts[bin_num][kmer] += 1
            else:
                kmer_counts[bin_num][kmer] = 1

            # Should we also do the mismatch?
            if (mismatch != 0):

                # Loop through all possible mismatches.
                for i_kmer in range(0, k):
                    for letter in alphabet:

                        # Don't count yourself as a mismatch.
                        if (kmer[i_kmer:i_kmer+1] != letter):

                            # Find the neighboring sequence.
                            neighbor = substitute(i_kmer, letter, kmer)

                            # If we're doing reverse complement, store the
                            # count in the version that starts with A or C.
                            if (revcomp == 1):
                                rev_kmer = find_revcomp(kmer,
                                                        revcomp_dictionary)
                                if (cmp(kmer, rev_kmer) > 0):
                                    kmer = rev_kmer

                            # Increment the neighboring sequence.
                            if (kmer_counts[bin_num].has_key(neighbor)):
                                kmer_counts[bin_num][neighbor] += mismatch
                            else:
                                kmer_counts[bin_num][neighbor] = mismatch

    # Build the sequence vector.
    sequence_vector = []
    for i_bin in range(0, num_bins):
        for kmer in kmer_list:
            #if (kmer_counts[i_bin].has_key(kmer)):
            if (kmer in kmer_counts[i_bin].keys()):				
                sequence_vector.append(kmer_counts[i_bin][kmer] + pseudocount)
            else:
                sequence_vector.append(pseudocount)

    # Normalize it
    return_value = normalize_vector(normalize_method,
                                    k_values,
                                    sequence_vector,
                                    kmer_list)

    return(return_value)

##############################################################################
def read_fasta_sequence (numeric,
                         fasta_file):

    # Read 1 byte.  
    first_char = fasta_file.read(1)
    # If it's empty, we're done.
    if (first_char == ""):
        return(["", ""])
    # If it's ">", then this is the first sequence in the file.
    elif (first_char == ">"):
        line = ""
    else:
        line = first_char

    # Read the rest of the header line.
    line = line + fasta_file.readline()
    # Get the rest of the ID.
    words = line.split()
    if (len(words) == 0):
      sys.stderr.write("No words in header line (%s)\n" % line)
      sys.exit(1)  
    id = words[0]
    #id = words[1].split("=")[1]
        
    # Read the sequence, through the next ">".
    first_char = fasta_file.read(1)
    sequence = ""
    while ((first_char != ">") and (first_char != "")):
        if (first_char != "\n"): # Handle blank lines.
            line = fasta_file.readline()
            sequence = sequence + first_char + line
        first_char = fasta_file.read(1)

    # Remove EOLs.
    clean_sequence = ""
    for letter in sequence:
        if (letter != "\n"):
            clean_sequence = clean_sequence + letter
    sequence = clean_sequence

    # Remove spaces.
    if (numeric == 0):
        clean_sequence = ""
        for letter in sequence:
            if (letter != " "):
                clean_sequence = clean_sequence + letter
        sequence = clean_sequence.upper()
        
    return([id, sequence])

##############################################################################
def read_sequence_and_numbers(fasta_file,
                              numbers_filename,
                              numbers_file):
    [fasta_id, fasta_sequence] = read_fasta_sequence(0, fasta_file)
    if (number_filename != ""):
        [number_id, number_sequence] = read_fasta_sequence(1, number_file)

        # Make sure we got the same ID.
        if (fasta_id != number_id):
            sys.stderr.write("Found mismatching IDs (%s != %d)\n" %
                             (fasta_id, number_id))
            sys.exit(1)

        # Split the numbers into a list.
        number_list = number_sequence.split()

        # Verify that they are the same length.
        if (len(fasta_sequence) != len(number_list)):
            sys.stderr.write("Found sequence of length %d with %d numbers.\n"
                             % (len(sequence), len(number_list)))
            print(sequence)
            print(numbers)
            sys.exit(1)
    else:
        number_list = ""

    return(fasta_id, fasta_sequence, number_list)
                                                      
    
##############################################################################
def find_revcomp (sequence,
                  revcomp_dictionary):

    # Save time by storing reverse complements in a hash.
    if (revcomp_dictionary.has_key(sequence)):
        return(revcomp_dictionary[sequence])

    # Make a reversed version of the string.
    rev_sequence = list(sequence)
    rev_sequence.reverse()
    rev_sequence = ''.join(rev_sequence)
    
    return_value = ""
    for letter in rev_sequence:
        if (letter == "A"):
            return_value = return_value + "T"
        elif (letter == "C"):
            return_value = return_value + "G"
        elif (letter == "G"):
            return_value = return_value + "C"
        elif (letter == "T"):
            return_value = return_value + "A"
        elif (letter == "N"):
            return_value = return_value + "N"
        else:
            sys.stderr.write("Unknown DNA character (%s)\n" % letter)
            sys.exit(1)

    # Store this value for future use.
    revcomp_dictionary[sequence] = return_value

    return(return_value)

##############################################################################
def compute_quantile_boundaries (num_bins,
                                 k_values,
                                 number_filename):

    if (num_bins == 1):
        return

    # The boundaries are stored in a 2D dictionary.
    boundaries = {}

    # Enumerate all values of k.
    for k in k_values:

        # Open the number file for reading.
        number_file = open(number_filename, "r")
        
        # Read it sequence by sequence.
        all_numbers = []
        [id, numbers] = read_fasta_sequence(1, number_file)
        while (id != ""):
            
            # Compute and store the mean of all k-mers.
            number_list = numbers.split()
            num_numbers = len(number_list) - k
            for i_number in range(0, num_numbers):
                if (i_number == 0):
                    sum = 0;
                    for i in range(0, k):
                        sum += float(number_list[i])
                else:
                    sum -= float(number_list[i_number - 1])
                    sum += float(number_list[i_number + k - 1])
                all_numbers.append(sum / k)
            [id, numbers] = read_fasta_sequence(1, number_file)
        number_file.close()

        # Sort them.
        all_numbers.sort()

        # Record the quantiles.
        boundaries[k] = {}
        num_values = len(all_numbers)
        bin_size = float(num_values) / float(num_bins)
        sys.stderr.write("boundaries k=%d:" % k)
        for i_bin in range(0, num_bins):
            value_index = int((bin_size * (i_bin + 1)) - 1)
            if (value_index == num_bins - 1):
                value_index = num_values - 1
            value = all_numbers[value_index]
            boundaries[k][i_bin] = value
            sys.stderr.write(" %g" % boundaries[k][i_bin])
        sys.stderr.write("\n")

    return(boundaries)
    

##############################################################################
# MAIN
##############################################################################

# Define the command line usage.
usage = """Usage: fasta2matrix [options] <k> <fasta file>

  Options:

    -upto       Use all values from 1 up to the specified k.

    -revcomp    Collapse reverse complement counts.

    -normalize [frequency|unitsphere] Normalize counts to be 
                frequencies or project onto unit sphere.  With -upto,
                normalization is done separately for each k.

    -protein    Use an amino acid alphabet.  Default=ACGT.

    -alphabet <string> Set the alphabet arbitrarily.

    -mismatch <value>  Assign count of <value> to k-mers that 
                       are 1 mismatch away.
    
    -binned <numbins> <file>  Create <numbins> vectors for each
                              sequence, and place each k-mer count
                              into the bin based upon its corresponding
                              mean value from the <file>.  The
                              <file> is in FASTA-like format, with
                              space-delimited numbers in place of
                              the sequences.  The sequences must
                              have the same names and be in the same
                              order as the given FASTA file.

   -pseudocount <value>  Assign the given pseudocount to each bin.

"""

# Define default options.
upto = 0
revcomp = 0
normalize_method = "none"
alphabet = "ACGT"
mismatch = 0
num_bins = 1
pseudocount = 0
number_filename = ""

# Parse the command line.
# sys.argv = sys.argv[1:]
# while (len(sys.argv) > 2):
#   next_arg = sys.argv[0]
#   sys.argv = sys.argv[1:]
#   if (next_arg == "-revcomp"):
#     revcomp = 1
#   elif (next_arg == "-upto"):
#     upto = 1
#   elif (next_arg == "-normalize"):
#     normalize_method = sys.argv[0]
#     sys.argv = sys.argv[1:]
#     if ((normalize_method != "unitsphere") and
#       (normalize_method != "frequency")):
#       sys.stderr.write("Invalid normalization method (%s).\n"
#                        % normalize_method);
#       sys.exit(1)
#   elif (next_arg == "-protein"):
#     alphabet = "ACDEFGHIKLMNPQRSTVWY"
#   elif (next_arg == "-alphabet"):
#     alphabet = sys.argv[0]
#     sys.argv = sys.argv[1:]
#   elif (next_arg == "-mismatch"):
#     mismatch = float(sys.argv[0])
#     sys.argv = sys.argv[1:]
#   elif (next_arg == "-binned"):
#     num_bins = int(sys.argv[0])
#     sys.argv = sys.argv[1:]
#     number_filename = sys.argv[0]
#     sys.argv = sys.argv[1:]
#   elif (next_arg == "-pseudocount"):
#     pseudocount = int(sys.argv[0])
#     sys.argv = sys.argv[1:]
#   else:
#     sys.stderr.write("Invalid option (%s)\n" % next_arg)
#     sys.exit(1)
# if (len(sys.argv) != 2):
#   sys.stderr.write(usage)
#   sys.exit(1)
k = int(sys.argv[1])
fasta_filename = sys.argv[2]
output_filename = sys.argv[3]
# Check for reverse complementing non-DNA alphabets.
if ((revcomp == 1) and (alphabet != "ACGT")):
  sys.stderr.write("Attempted to reverse complement ")
  sys.stderr.write("a non-DNA alphabet (%s)\n" % alphabet)

# Make a list of all values of k.
k_values = []
if (upto == 1):
  start_i_k = 1
else:
  start_i_k = k
k_values = range(start_i_k, k+1)

# If numeric binning is turned on, compute quantile boundaries for various
# values of k.
boundaries = compute_quantile_boundaries(num_bins, k_values, number_filename)
  
# Make a list of all k-mers.
kmer_list = make_upto_kmer_list(k_values, alphabet);
sys.stdout.write("Consideriamo %d kmers.\n" % len(kmer_list))

# Set up a dictionary to cache reverse complements.
revcomp_dictionary = {}

# Use lexicographically first version of {kmer, revcomp(kmer)}.
if (revcomp == 1):
  new_kmer_list = []
  for kmer in kmer_list:
      rev_kmer = find_revcomp(kmer, revcomp_dictionary)
      if (cmp(kmer, rev_kmer) <= 0):
          new_kmer_list.append(kmer)
  kmer_list = new_kmer_list;
  sys.stdout.write("Reduced to %d kmers.\n" % len(kmer_list))

# Print the corner of the matrix.

outfile=open(output_filename,"w")

# Print the title row.


sys.stdout.write("\n")
outfile.write("seq_id,")

# Open the sequence file.
if (fasta_filename == "-"):
  fasta_file = sys.stdin
else:
  fasta_file = open(fasta_filename, "r")
if (number_filename == ""):
  number_file = 0
else:
  number_file = open(number_filename, "r")
#for i_bin in range(1, num_bins+1):
for kmer in kmer_list:
  #if (num_bins > 1):
    #outfile.write("%s-%d," % (kmer, i_bin))
      #i+=1
  #else:
  if(kmer==kmer_list[len(kmer_list)-1]):
      outfile.write("%s" % kmer)
  else:
	  outfile.write("%s," %kmer)
      #i+=1

outfile.write("\n")
# Read the first sequence.
[id, sequence, numbers] = read_sequence_and_numbers(fasta_file,
                                                    number_filename,
                                                    number_file)
# Iterate till we've read the whole file.
i_sequence = 1
vett=np.zeros(len(kmer_list),dtype=int)
while (id != ""):

  # Tell the user what's happening.
  if (i_sequence % 1000 == 0):
    sys.stdout.write("Reading %dth sequenza.\n" % i_sequence)

  # Compute the sequence vector.
  vector = make_sequence_vector(sequence,
                                numbers,
                                num_bins,
                                revcomp,
                                revcomp_dictionary,
                                normalize_method,
                                k_values,
                                mismatch,
                                alphabet,
                                kmer_list,
                                boundaries,
                                pseudocount)

  # Print the formatted vector.
  outfile.write("%s," % id)
  
  count=len(kmer_list)

  for element in vector:
    if(count!=1):
        outfile.write("%d," % element)
    else:
        outfile.write("%d" % element)
    count = count-1
        
  outfile.write("\n")
  # Read the next sequence.
  [id, sequence, numbers] = read_sequence_and_numbers(fasta_file,
                                                      number_filename,
                                                      number_file)
  i_sequence += 1
# Close the file.
  i=0


outfile.close()
fasta_file.close()








#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 19 11:56:08 2021

@author: duong
"""
import sys
if sys.version_info[0] >= 3:
	unicode = str
import json
import os, argparse
from Bio import SeqIO
import random
import multiprocessing
parser=argparse.ArgumentParser(prog='compareClassification.py',  
							   usage="%(prog)s [options] -c1 cbsclassification1 -c2 classification2 -o output",
							   description='''Script that compares two classifications.''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-c1','--classificationfilename1', required=True, help='the classification filename 1')
parser.add_argument('-c2','--classificationfilename2', required=True, help='the classification filename 2')
parser.add_argument('-o','--out', required=True, help='The output.')

args=parser.parse_args()
classificationfilename1= args.classificationfilename1
classificationfilename2= args.classificationfilename2
output=args.out

def LoadClassificationAtPos(classificationfilename,pos,p_score):
	classificationfile=open(classificationfilename)
	next(classificationfile)
	taxa={}
	for line in classificationfile:
		texts=line.rstrip().split("\t")
		seqid=texts[0]
		taxonname=""
		score=0
		if pos <len(texts):
			taxonname=texts[pos]
			taxonname=taxonname.replace("_"," ")
			if ("uncultured" in taxonname) or ("unidentified" in taxonname) or (" sp " in taxonname) :
				taxonname="unidentified"
		if p_score <len(texts):
			score=float(texts[p_score])
		if taxonname=="" or taxonname=="unidentified":
			continue
		taxa.setdefault(seqid,[taxonname,score])
	classificationfile.close()
	return taxa

def LoadClassification(classificationfilename):
	kingdoms={}
	phyla={}
	classes={}
	orders={}
	families={}
	genera={}
	species={}
	classificationfile=open(classificationfilename)
	header=next(classificationfile)
	classificationfile.close()
	p_s=-1
	p_g=-1
	p_f=-1
	p_o=-1
	p_c=-1
	p_p=-1
	p_k=-1
	p_score=-1
	i=0
	for text in header.split("\t"):
		text=text.rstrip()
		if text.lower()=="species":
			p_s=i
		elif text.lower()=="genus":
			p_g=i	
		elif text.lower()=="family":
			p_f=i	
		elif text.lower()=="order":
			p_o=i	
		elif text.lower()=="class":
			p_c=i	
		elif text.lower()=="phylum":
			p_p=i	
		elif text.lower()=="kingdom":
			p_k=i
		elif text.lower()=="score":
			p_score=i	
		i=i+1 
	species=LoadClassificationAtPos(classificationfilename,p_s,p_score)	
	genera=LoadClassificationAtPos(classificationfilename,p_g,p_score)	
	families=LoadClassificationAtPos(classificationfilename,p_f,p_score)	
	orders=LoadClassificationAtPos(classificationfilename,p_o,p_score)	
	classes=LoadClassificationAtPos(classificationfilename,p_c,p_score)		
	phyla=LoadClassificationAtPos(classificationfilename,p_p,p_score)	
	kingdoms=LoadClassificationAtPos(classificationfilename,p_k,p_score)	
	return kingdoms,phyla,classes,orders,families,genera,species, p_k,p_p,p_c,p_o,p_f,p_g,p_s

def Compare(rank,classes1,classes2,classification1,classification2,output):
	outputfile=open(output,"a")
	taxafile=open(output + "." + rank,"w")
	number1=len(classes1.keys())
	number2=len(classes2.keys())
	seqids=list(set(classes1.keys()) & set(classes2.keys()))
	number=len(seqids)
	taxafile.write("Sequence ID\tClassification of " + classification1 + "\tScore given in " + classification1 + "\tClassification of" + classification2 + "\tScore given in " + classification2 +"\tCompared" + "\n")
	same=0
	badseqids1=[]
	badseqids2=[]
	seqids1=[]
	seqids2=[]
	onlyseqids1=[]
	onlyseqids2=[]
	notonlyseqids1=[]
	notonlyseqids2=[]
	count=0
	for seqid in classes1.keys():
		taxon1=classes1[seqid][0]
		score1=classes1[seqid][1]
		taxon2=""
		score2=0
		compared=""	
		if seqid in classes2.keys():
			taxon2=classes2[seqid][0]
			score2=classes2[seqid][1]
			if taxon1==taxon2:
				same=same+1
				compared="same"
				seqids1.append(seqid)	
				seqids2.append(seqid)	
				notonlyseqids1.append(seqid)	
				notonlyseqids2.append(seqid)	
			else:
				compared="diff"
				if score1>=score2:
					onlyseqids1.append(seqid)	
					seqids1.append(seqid)	
				if score2>=score1:
					onlyseqids2.append(seqid)	
					seqids2.append(seqid)		
				if score1<score2:
					badseqids1.append(seqid)
					notonlyseqids1.append(seqid)	
				elif score2<score1:
					count=count+1
					badseqids2.append(seqid)
					notonlyseqids2.append(seqid)	
		else:
			onlyseqids1.append(seqid)	
			seqids1.append(seqid)	
		taxafile.write(seqid + "\t" + taxon1 + "\t" +str(score1) + "\t" + taxon2 + "\t" + str(score2) + "\t" + compared + "\n")
	for seqid in classes2.keys():
		if seqid in classes1.keys():
			continue
		onlyseqids2.append(seqid)	
		seqids2.append(seqid)	
		taxon1=""
		score1=0
		taxon2=classes2[seqid][0]
		score2=classes2[seqid][1]
		compared=""	
		taxafile.write(seqid + "\t" + taxon1 + "\t" + str(score1) + "\t" + taxon2 + "\t" + str(score2) +"\t" + compared + "\n")	
	taxafile.close()	
	difference=number-same
	onlyby1=number1-number
	onlyby2=number2-number
	higherscoreby1=count
	lessscoreby1=difference-count
	higherscoreby2=difference-count
	lessscoreby2=count
	outputfile.write(rank + "\t" + str(number1)+ "\t" + str(number2)+ "\t" + str(number) + "\t" + str(same) + "\t" + str(difference) + "\t" + str(higherscoreby1) + "\t" + str(higherscoreby2) + "\t" + str(lessscoreby1) + "\t" + str(lessscoreby2) + "\t" + str(onlyby1) + "\t" + str(onlyby2) + "\n" )
	outputfile.close()
	print(rank + "\t" + str(number1)+ "\t" + str(number2)+ "\t" + str(number) + "\t" + str(same) + "\t" + str(difference) + "\t" + str(higherscoreby1) + "\t" + str(higherscoreby2) + "\t" + str(lessscoreby1) + "\t" + str(lessscoreby2) + "\t" + str(onlyby1) + "\t" + str(onlyby2) + "\n" )
	print("The comparison at the " + rank + " taxonomic level is saved in file " +  output + "." + rank + "." )
	return seqids1,seqids2,onlyseqids1,onlyseqids2
	
def CreateNewClassificationFile(filename,newfilename,skeys,gkeys,fkeys,okeys,ckeys,pkeys,kkeys,p_k,p_p,p_c,p_o,p_f,p_g,p_s):
	newfile=open(newfilename,"w")
	file=open(filename)
	header=next(file)
	newfile.write(header)
	for line in file:
		newline=""
		seqid=line.split("\t")[0]
		texts=line.split("\t")
		i=0
		for text in texts:
			text=text.rstrip()
			if i==p_s:
				if not (seqid in skeys):
					text="unidentified"
			if i==p_g:
				if not (seqid in gkeys):
					text="unidentified"		
			if i==p_f:
				if not (seqid in fkeys):
					text="unidentified"	
			if i==p_o:
				if not (seqid in okeys):
					text="unidentified"		
			if i==p_c:
				if not (seqid in ckeys):
					text="unidentified"	
			if i==p_p:
				if not (seqid in pkeys):
					text="unidentified"			
			if i==p_k:
				if not (seqid in kkeys):
					text="unidentified"			
			newline=newline + text +"\t" 		
			i=i+1
		newline=newline[:-1] + "\n"	
		newfile.write(newline)	
	newfile.close()
	file.close()
	
def CreateNewClassificationFileForSelectedSeqIds(filename,newfilename,skeys,gkeys,fkeys,okeys,ckeys,pkeys,kkeys):
	newfile=open(newfilename,"w")
	file=open(filename)
	header=next(file)
	newfile.write(header)
	for line in file:
		seqid=line.split("\t")[0]
		isok=0
		if (seqid in skeys):
			isok=1
		elif seqid in gkeys:
			isok=1
		elif seqid in fkeys:
			isok=1
		elif seqid in okeys:
			isok=1
		elif seqid in ckeys:
			isok=1
		elif seqid in pkeys:
			isok=1
		elif seqid in kkeys:
			isok=1
		if isok==1:
			newfile.write(line)	
	newfile.close()
	file.close()	
	
#def CreateNewClassificationFileForSelectedSeqIds(filename,newfilename,skeys,selectedsseqids,gkeys,selectedgseqids,fkeys,selectedfseqids,okeys,selectedoseqids,ckeys,selectedcseqids,pkeys,selectedpseqids,kkeys,selectedkseqids):
#	newfile=open(newfilename,"w")
#	file=open(filename)
#	header=next(file)
#	newfile.write(header)
#	for line in file:
#		seqid=line.split("\t")[0]
#		isok=0
#		if (seqid in skeys):
#			if (seqid in selectedsseqids):
#				isok=1
#		elif (seqid in gkeys):
#			if seqid in selectedgseqids:
#				isok=1
#		elif (seqid in fkeys):
#			if (seqid in selectedfseqids):
#				isok=1
#		elif (seqid in okeys):
#			if (seqid in selectedoseqids):
#				isok=1
#		elif (seqid in ckeys):
#			if (seqid in selectedcseqids):
#				isok=1
#		elif (seqid in pkeys):
#			if (seqid in selectedpseqids):
#				isok=1
#		elif (seqid in kkeys):
#			if (seqid in selectedkseqids):
#				isok=1
#		if isok==1:
#			newfile.write(line)	
#	newfile.close()
#	file.close()	
###

classification1=os.path.basename(classificationfilename1)
classification2=os.path.basename(classificationfilename2)
kingdoms1,phyla1,classes1,orders1,families1,genera1,species1,p_k1,p_p1,p_c1,p_o1,p_f1,p_g1,p_s1=LoadClassification(classificationfilename1)
kingdoms2,phyla2,classes2,orders2,families2,genera2,species2,p_k2,p_p2,p_c2,p_o2,p_f2,p_g2,p_s2=LoadClassification(classificationfilename2)
outputfile=open(output,"w")	
outputfile.write("Rank\tSequence number of " + classification1+ "\tSequence number of " + classification2 + "\tNumber of common sequences\tNumber of the same classification\tNumber of different classifications\tHigher scores by first classification\tHigher scores by second classification\tLess scores by first classification\tLess scores by second classification\tOnly by first classification\tOnly by second classification\n")
outputfile.close()
sseqids1,sseqids2,onlysseqids1,onlysseqids2=Compare("species",species1,species2,classification1,classification2,output)
gseqids1,gseqids2,onlygseqids1,onlygseqids2=Compare("genus",genera1,genera2,classification1,classification2,output)
fseqids1,fseqids2,onlyfseqids1,onlyfseqids2=Compare("family",families1,families2,classification1,classification2,output)
oseqids1,oseqids2,onlyoseqids1,onlyoseqids2=Compare("order",orders1,orders2,classification1,classification2,output)
cseqids1,cseqids2,onlycseqids1,onlycseqids2=Compare("class",classes1,classes2,classification1,classification2,output)
pseqids1,pseqids2,onlypseqids1,onlypseqids2=Compare("phylum",phyla1,phyla2,classification1,classification2,output)
kseqids1,kseqids2,onlykseqids1,onlykseqids2=Compare("kingdom",kingdoms1,kingdoms2,classification1,classification2,output)
outputfile.close()
#write better classification to file
newclassificationfilename1=classificationfilename1 + ".worseclassificationremoved"
newclassificationfilename2=classificationfilename2 + ".worseclassificationremoved"
CreateNewClassificationFile(classificationfilename1,newclassificationfilename1,sseqids1,gseqids1,fseqids1,oseqids1,cseqids1,pseqids1,kseqids1,p_k1,p_p1,p_c1,p_o1,p_f1,p_g1,p_s1)
CreateNewClassificationFile(classificationfilename2,newclassificationfilename2,sseqids2,gseqids2,fseqids2,oseqids2,cseqids2,pseqids2,kseqids2,p_k2,p_p2,p_c2,p_o2,p_f2,p_g2,p_s2)

##write only classification to file
newclassificationfilename1_by_1_only=classificationfilename1 + ".only"
newclassificationfilename2_by_2_only=classificationfilename2 + ".only"
CreateNewClassificationFileForSelectedSeqIds(classificationfilename1,newclassificationfilename1_by_1_only,onlysseqids1,onlygseqids1,onlyfseqids1,onlyoseqids1,onlycseqids1,onlypseqids1,onlykseqids1)
CreateNewClassificationFileForSelectedSeqIds(classificationfilename2,newclassificationfilename2_by_2_only,onlysseqids2,onlygseqids2,onlyfseqids2,onlyoseqids2,onlycseqids2,onlypseqids2,onlykseqids2)

print("The comparison is saved in file " + output + ".")
print("The classifications with worse classifications removed are saved in files " + newclassificationfilename1 + " and " + newclassificationfilename2 + ".")
print("The only classifications are saved in files " + newclassificationfilename1_by_1_only + " and " + newclassificationfilename2_by_2_only + ".")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Apr 17 13:34:05 2021

@author: duong
"""
import sys
import os, argparse
from Bio import SeqIO

parser=argparse.ArgumentParser(prog='cleanClassification.py',  
							   usage="%(prog)s [options] -i classificationfilename -o outputname",
							   description='''The script that clean classification adding missing classification to the sequences. ''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )


parser.add_argument('-i','--input', required=True, help='The taxonomic classification file.')
parser.add_argument('-o','--out',  help='The output folder.')

args=parser.parse_args()
classificationfilename= args.input
output=args.out

def GetTaxonomicClassification(classificationpos,header,texts):
	classification=""
	p_s=len(texts)
	p_g=len(texts)
	p_f=len(texts)
	p_o=len(texts)
	p_c=len(texts)
	p_p=len(texts)
	p_k=len(texts)
	i=0
	for text in header.split("\t"):
		text=text.rstrip()
		if text.lower()=="species":
			p_s=i
		elif text.lower()=="genus":
			p_g=i	
		elif text.lower()=="family":
			p_f=i	
		elif text.lower()=="order":
			p_o=i	
		elif text.lower()=="class":
			p_c=i	
		elif text.lower()=="phylum":
			p_p=i	
		elif text.lower()=="kingdom":
			p_k=i	
		i=i+1 
	species=""
	genus=""
	family=""
	order="" 
	bioclass=""
	phylum=""
	kingdom=""
	if p_s< len(texts):
		species=texts[p_s].rstrip()
	if p_g< len(texts):
		genus=texts[p_g].rstrip()
	if p_f< len(texts):
		family=texts[p_f].rstrip()
	if p_o< len(texts):
		order=texts[p_o].rstrip()
	if p_c< len(texts):
		bioclass=texts[p_c].rstrip()
	if p_p< len(texts):
		phylum=texts[p_p].rstrip()
	if p_k< len(texts):
		kingdom=texts[p_k].rstrip()
	if classificationpos==0:
		classificationpos=p_s
	level=7	
	if classificationpos==p_s:	
		level=0
	elif classificationpos==p_g:	
		level=1
	elif classificationpos==p_f:	
		level=2
	elif classificationpos==p_o:	
		level=3
	elif classificationpos==p_c:	
		level=4
	elif classificationpos==p_p:	
		level=5
	elif classificationpos==p_k:	
		level=6
	taxonname=""
	if classificationpos <len(texts):
		taxonname=texts[classificationpos]		
	if level <7 and kingdom!="":
		taxonname=kingdom
		rank="kingdom"
		classification=kingdom
	if level <6 and phylum!="":
		taxonname=phylum
		rank="phylum"
		classification=kingdom +";"+phylum 
	if level <5 and bioclass!="":
		taxonname=bioclass
		rank="class"
		classification=kingdom +";"+phylum +";"+bioclass
	if level <4 and order!="":
		taxonname=order
		rank="order"
		classification=kingdom +";"+phylum +";"+bioclass +";"+order
	if level <3 and family!="":
		taxonname=family
		rank="family"
		classification=kingdom +";"+phylum +";"+bioclass +";"+order+";"+family 
	if level <2 and genus!="":
		taxonname=genus
		rank="genus"
		classification=kingdom +";"+phylum +";"+bioclass +";"+order+";"+family + ";"+ genus
	if level <1 and species!="":
		taxonname=species
		rank="species"
		classification=kingdom +";"+phylum +";"+bioclass +";"+order+";"+family + ";"+ genus+";"+species 	
	return classification,taxonname,rank	

def LoadClassification(classificationfilename,header,pos):
	classificationdict={}
	classificationfile= open(classificationfilename)
	next(classificationfile)
	for line in classificationfile:
		texts=line.split("\t")
		if len(texts) < len(header.split("\t")):
			continue
		#only consider the good format
		if pos < len(texts):
			classname=texts[pos]
			if classname!="":
				classification,classname,rank=GetTaxonomicClassification(pos,header,texts)
				if classname in classificationdict.keys():
					if len(classificationdict[classname].split("\t")) < len(classification.split("\t")):
						print(line)
						print(classificationdict[classname])
						print(classification)
						classificationdict[classname]=classification
				else:
					classificationdict.setdefault(classname,classification)
	classificationfile.close()	
	return classificationdict



##############MAIN############################
classificationfile=open(classificationfilename)
headers=next(classificationfile)
classificationfile.close()
texts=headers.rstrip().split("\t")
p_s=len(texts)
p_g=len(texts)
p_f=len(texts)
p_o=len(texts)
p_c=len(texts)
p_p=len(texts)
p_k=len(texts)
i=0
for text in headers.split("\t"):
	text=text.rstrip()
	if text.lower()=="species":
		p_s=i
	elif text.lower()=="genus":
		p_g=i	
	elif text.lower()=="family":
		p_f=i	
	elif text.lower()=="order":
		p_o=i	
	elif text.lower()=="class":
		p_c=i	
	elif text.lower()=="phylum":
		p_p=i	
	elif text.lower()=="kingdom":
		p_k=i	
	i=i+1 
	
speciesdict=LoadClassification(classificationfilename,headers,p_s)
genusdict=LoadClassification(classificationfilename,headers,p_g)
familydict=LoadClassification(classificationfilename,headers,p_f)
orderdict=LoadClassification(classificationfilename,headers,p_o)
classdict=LoadClassification(classificationfilename,headers,p_c)
phylumdict=LoadClassification(classificationfilename,headers,p_p)
kingdomdict=LoadClassification(classificationfilename,headers,p_k)

classificationfile=open(classificationfilename)
next(classificationfile)
outputfile=open(output,"w")	
outputfile.write(headers)
for line in classificationfile:
	texts=line.split("\t")
	species=""
	genus=""
	family=""
	order=""
	bioclass=""
	phylum=""
	kingdom=""
	for i in range(0,len(texts)):
		text=texts[len(texts)-i-1].rstrip()
		if text in speciesdict.keys():
			classificationtext=speciesdict[text]
			classification=classificationtext.split(";")
			if species=="":
				species = text
			if genus=="":	
				genus=classification[5]
			if family=="":	
				family=classification[4]
			if order=="":	
				order=classification[3]
			if bioclass=="":	
				bioclass=classification[2]
			if phylum=="":
				phylum=classification[1]
			if kingdom=="":
				kingdom=classification[0]
		elif text in genusdict.keys():
			classificationtext=genusdict[text]
			classification=classificationtext.split(";")
			if genus=="":	
				genus=text
			if family=="":	
				family=classification[4]
			if order=="":	
				order=classification[3]
			if bioclass=="":	
				bioclass=classification[2]
			if phylum=="":
				phylum=classification[1]
			if kingdom=="":
				kingdom=classification[0]
		elif text in familydict.keys():
			classificationtext=familydict[text]
			classification=classificationtext.split(";")
			if family=="":	
				family=text
			if order=="":	
				order=classification[3]
			if bioclass=="":	
				bioclass=classification[2]
			if phylum=="":
				phylum=classification[1]
			if kingdom=="":
				kingdom=classification[0]
		elif text in orderdict.keys():	
			classificationtext=orderdict[text]
			classification=classificationtext.split(";")
			if order=="":	
				order=text
			if bioclass=="":	
				bioclass=classification[2]
			if phylum=="":
				phylum=classification[1]
			if kingdom=="":
				kingdom=classification[0]
		elif text in classdict.keys():	
			classificationtext=classdict[text]
			classification=classificationtext.split(";")
			if bioclass=="":	
				bioclass=text
			if phylum=="":
				phylum=classification[1]
			if kingdom=="":
				kingdom=classification[0]
		elif text in phylumdict.keys():	
			classificationtext=phylumdict[text]
			classification=classificationtext.split(";")
			if phylum=="":
				phylum=text
			if kingdom=="":
				kingdom=classification[0]		
	i=0		
	newclassification=""
	for i in range(0,len(headers.split("\t"))):
		text=""
		if i <len(texts):
			text=texts[i].rstrip()
		if i==p_s:
			text=species
		if i==p_g:
			text=genus
		if i==p_f:
			text=family
		if i==p_o:
			text=order
		if i==p_c:
			text=bioclass
		if i==p_p:
			text=phylum
		if i==0:
			newclassification=text
		else:
			newclassification=newclassification + "\t" + text
		i=i+1	
	outputfile.write(newclassification+"\n")
outputfile.close()


#!/usr/bin/env python
import sys
if sys.version_info[0] >= 3:
	unicode = str
import os, argparse
from Bio import SeqIO
from Bio import Phylo
import pylab
import multiprocessing
nproc=multiprocessing.cpu_count()

parser=argparse.ArgumentParser(prog='maketree.py',  
							   usage="%(prog)s [options] -i fastafile -c classificationfilename -rank classification rank",
							   description='''The script create a phylogenetic tree for the sequences given in the input fasta file. ''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--input', required=True, help='the fasta input file.')
parser.add_argument('-c','--classification', default="", help='The taxonomic classification file.')
#parser.add_argument('-p','--classificationposition', default="", help='the classification positions for the prediction.')
parser.add_argument('-ranks','--classificationranks', default="", help='the classification ranks for getting sequence dscriptions.')
parser.add_argument('-o','--out', default="dnabarcoder", help='The output folder.')

args=parser.parse_args()
fastafilename= args.input
classificationfilename= args.classification
ranks=args.classificationranks
outputpath=args.out

outputpath=args.out
if not os.path.exists(outputpath):
	os.system("mkdir " + outputpath)	

def GetBase(filename):
	return filename[:-(len(filename)-filename.rindex("."))]

def GetWorkingBase(filename):
	basename=os.path.basename(filename)
	basename=basename[:-(len(basename)-basename.rindex("."))] 
	path=outputpath + "/" + basename
	return path

def LoadClassification(classificationfilename,poslist):
	classificationfile= open(classificationfilename, errors="ignore")
	seqids=[]
	classifications=[]
	numberoffeatures=0
	for line in classificationfile:
		elements=line.rstrip().split("\t")
		seqid = elements[0].replace(">","").rstrip()
		seqids.append(seqid)
		classification=(line[line.index("\t")+1:])
		texts=classification.split("\t")
		classification=""
		if len(poslist) >0:		
			for pos in poslist:
				pos=pos-1
				text=""
				if pos < len(texts):
					text=texts[pos]
				if text=="":
					text="unidentified"	
				text=text.replace(" ","_")	
				classification=classification + text + "|"
		else:
			for text in texts:
				text=text.rstrip()
				if text=="":
					text="unidentified"
				text=text.replace(" ","_")		
				classification=classification + text + "|"
		classification=classification[:-1] 		
		classifications.append(classification)
		if numberoffeatures==0:
			numberoffeatures=classification.count("|")
		else:
			numberoffeatures=min(numberoffeatures,classification.count("|"))
	classificationfile.close()
	return seqids,classifications,numberoffeatures

def lookup_by_names(tree):
	names = {}
	for clade in tree.find_clades():
		if clade.name:
			if clade.name in names:
				raise ValueError("Duplicate key: %s" % clade.name)
			names[clade.name] = clade
	return names
#
#
def computeBranchLength(treefilename):
	tree = Phylo.read(treefilename, "newick")
	names = lookup_by_names(tree)
	for name in names:
		clade=names[name]
		print(clade.name + " " + str(clade.branch_length))


def PrintTree_ete(treefilename):
	svgfilename=GetWorkingBase(fastafilename) + ".tree.svg"
	from ete3 import Tree, TreeStyle, NodeStyle
	ts = TreeStyle()
	ts.show_leaf_name = True
	ts.show_branch_length = False
	ts.show_branch_support = True

	ns = NodeStyle()
	ns["size"] = 0
	ns["vt_line_width"] = 2
	ns["hz_line_width"] = 2

	# read tree
	t = Tree(treefilename)
	t.ladderize(1)
	for node in t.traverse():
		node.set_style(ns)
	# print final tree
	t.render(svgfilename, tree_style=ts)
	print("A iq-tree and its svg file are saved in " + treefilename + " and " + svgfilename + ".")	
	
def PrintTree(treefilename):
	svgfilename=GetWorkingBase(fastafilename) + ".tree.svg"
	tree = Phylo.read(treefilename, "newick")
	Phylo.draw(tree,do_show=True)	
	#Phylo.draw_ascii(tree,do_show=False)	
	pylab.savefig(svgfilename,format='svg', bbox_inches='tight', dpi=300)
	print("A iq-tree and its svg file are saved in " + treefilename + " and " + svgfilename + ".")	
	
def CreateTree(fastafilename):
	alignmentfilename =  GetWorkingBase(fastafilename) + ".aligned.fas"
	#make the alignment
	if not os.path.exists(alignmentfilename):
		command="clustalo -i " + fastafilename + " -o " + alignmentfilename
		print(command)
		os.system(command)
	#make tree
	treefilename=GetWorkingBase(fastafilename) + ".aligned.fas.treefile"
	if not os.path.exists(treefilename):
		#command="iqtree -pers 0.2 -s " + alignmentfilename
		command="iqtree -pers 0.2 -n 500 -s " + alignmentfilename
		#command="iqtree -s " + alignmentfilename
		#command="fasttree -nt -quote " + alignmentfilename + " > " + treefilename
		print(command)
		os.system(command)
	#print tree
	PrintTree(treefilename)
	return treefilename
def CreateFastaFileWithClassification(fastafilename,classificationfilename,positionlist):
	newfastafilename=GetWorkingBase(fastafilename) + ".classification.fasta"
	seqids,classifications,numberoffeatures=LoadClassification(classificationfilename,positionlist)
	outputfile=open(newfastafilename,"w")	
	fastafile=open(fastafilename)
	for line in fastafile:
		if line.startswith(">"):
			seqid=line.rstrip().replace(">","")
			if "|" in seqid:
				seqid=seqid[0:seqid.index("|")]
			classification=""
			if seqid in seqids:
				classification=unicode(classifications[seqids.index(seqid)])
			else:
				classification="|" * numberoffeatures
			header=">" + seqid + "|" + classification + "\n"
			outputfile.write(header)	
		else:		
			outputfile.write(line)
	outputfile.close()		
	fastafile.close()
	return newfastafilename

def GetPositionList(classificationfilename,ranks):
	ranklist=[]	
	if "," in ranks:
		ranklist=ranks.split(",")
	elif ranks !="":
		ranklist.append(ranks)
	positionlist=[]
	classificationfile=open(classificationfilename)
	header=classificationfile.readline()
	header=header.rstrip()
	classificationfile.close()
	texts=header.split("\t")
	isError=False
	for rank in ranklist:
		if rank in texts:
			pos=texts.index(rank)
			positionlist.append(pos)
		else:
			print("The rank " + rank + " is not given in the classification." )
			isError=True
	return positionlist,ranklist,isError

######MAIN
positionlist,ranklist,isError=GetPositionList(classificationfilename,ranks)
if isError==True:
	sys.exit()
newfastafilename=fastafilename
if os.path.exists(classificationfilename):
	newfastafilename=CreateFastaFileWithClassification(fastafilename,classificationfilename,positionlist)
treefilename=CreateTree(newfastafilename)

